import 'dart:async';

/// The error returned by Context.err() when the context is canceled.
final exceptionCanceled = Exception("canceled");

/// The error returned by Context.err() when the context's deadline passes.
final exceptionDeadlineExceeded = Exception("deadline exceeded");

abstract class Context {
  /// create context with value or values
  ///
  /// when param is [Map], will be used as ZoneValues directly.
  /// when param is [List], each item will be convert to Map, when item is not [MapEntry], `runtimeType` of item will be the key.
  /// when param is [MapEntry], zoneValues will be with the [MapEntry].
  /// when param is not [MapEntry], `runtimeType` of param will be the key.
  static Context withValue(dynamic valueOrListOrMap) {
    Map zoneValues = {};

    if (valueOrListOrMap is List) {
      zoneValues = valueOrListOrMap.asMap().map(
            (key, value) =>
                value is MapEntry ? value : MapEntry(value.runtimeType, value),
          );
    } else if (valueOrListOrMap is Map) {
      zoneValues = valueOrListOrMap;
    } else if (valueOrListOrMap is MapEntry) {
      zoneValues = {valueOrListOrMap.key: valueOrListOrMap.value};
    } else {
      zoneValues = {valueOrListOrMap.runtimeType: valueOrListOrMap};
    }

    return _ValuesContext(zoneValues);
  }

  /// returns the value associated with this context for key or matched type
  ///
  /// ```dart
  /// var ctx = Context.withValue(
  ///     MapEntry("comparable-key", "value will put into context"),
  /// );
  ///
  /// ctx.run(() {
  ///   Context.value("comparable-key")
  /// });
  /// ```
  static T? value<T>([dynamic key]) {
    return Zone.current[key ??= T];
  }

  /// returns a Stream that's closed when work done on behalf of this context should be canceled.
  static Stream<Exception>? done() {
    return Zone.current[CancelableContext]?.done();
  }

  static CancelableContext withCancel() {
    return _CancelContext();
  }

  /// Executes [action] in this context.
  R run<R>(R Function() action);

  static CancelableContext withTimeout(Duration timeout) {
    return _TimerContext(DateTime.now().add(timeout));
  }

  static CancelableContext withDeadline(DateTime deadline) {
    return _TimerContext(deadline);
  }
}

abstract class CancelableContext extends Context {
  /// cancel context may with Exception
  ///
  /// the Exception notice to Stream's subscription from Context.done();
  /// the default Exception is exceptionCanceled.
  cancel([Exception? reason]);

  /// returns a Stream that's closed when work done on behalf of this context should be canceled.
  Stream<Exception>? done();
}

class _ValuesContext implements Context {
  final Map _zoneValues;

  _ValuesContext(this._zoneValues);

  @override
  R run<R>(R Function() action) {
    return runZoned<R>(action, zoneValues: _zoneValues);
  }
}

class _CancelContext with _CancelableContext {
  _CancelContext() {
    _unsubscribe ??= _subscribeCancelsIfExists(cancel: _cancel);
  }
}

mixin _CancelableContext implements CancelableContext {
  Function()? _unsubscribe;
  Exception? _exception;
  final StreamController<Exception> _done = StreamController.broadcast();

  @override
  cancel([Exception? reason]) {
    _cancel(reason ?? exceptionCanceled);
    if (_unsubscribe != null) {
      _unsubscribe!();
      _unsubscribe = null;
    }
  }

  _cancel(Exception reason) {
    if (_exception != null) {
      return;
    }
    _exception = reason;
    _done.sink.add(reason);
    _done.close();
  }

  @override
  R run<R>(R Function() action) {
    return runZonedGuarded<R>(
      action,
      (e, s) => cancel(e is Exception ? e : Exception(e)),
      zoneValues: {CancelableContext: this},
    )!;
  }

  @override
  Stream<Exception>? done() {
    return _done.stream
        .transform(StreamTransformer.fromBind(_throwExceptionIfExists));
  }

  Stream<Exception> _throwExceptionIfExists(Stream<Exception> stream) {
    var controller = StreamController<Exception>();

    StreamSubscription? subscription;

    throwExceptionOnce(Exception e) {
      controller.sink.add(e);
      controller.close();
    }

    controller.onListen = () {
      if (_exception != null) {
        throwExceptionOnce(_exception!);
        return;
      }

      subscription = stream.listen(throwExceptionOnce);
    };

    controller.onCancel = () {
      subscription?.cancel();
      subscription = null;
    };

    return controller.stream;
  }
}

class _TimerContext with _CancelableContext {
  _TimerContext(this._deadline) {
    _unsubscribe ??= _subscribeCancelsIfExists(
      cancel: _cancel,
      cancel$: Stream.periodic(_deadline.difference(DateTime.now()))
          .map((_) => exceptionDeadlineExceeded),
    );
  }

  final DateTime _deadline;
}

_subscribeCancelsIfExists({
  required Function(Exception) cancel,
  Stream<Exception>? cancel$,
}) {
  var parentCancel$ = Context.done();

  List<Future<Exception>> cancelFutures = [
    parentCancel$?.first,
    cancel$?.first,
  ].whereType<Future<Exception>>().toList();

  if (cancelFutures.isEmpty) {
    return () => {};
  }

  var sub = Future.any(cancelFutures).asStream().listen(
        cancel,
        onError: (err) => err is Exception ? cancel(err) : null,
      );

  return () => sub.cancel();
}
